/* Copyright (c) Mark J. Kilgard, 1994, 1995, 1996. */

/* This program is freely distributable without licensing fees
   and is provided without guarantee or warrantee expressed or
   implied. This program is -not- in the public domain. */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>

#include <GL/glut.h>
#include "glutint.h"
#include "GL/fxmesa.h"

static GLUTtimer *freeTimerList = NULL;
static int mappedMenuButton;

GLUTidleCB __glutIdleFunc = NULL;
GLUTtimer *__glutTimerList = NULL;
#ifdef SUPPORT_FORTRAN
GLUTtimer *__glutNewTimer;
#endif
GLUTwindow *__glutWindowWorkList = NULL;
void (*__glutUpdateInputDeviceMaskFunc) (GLUTwindow *);
Atom __glutMotifHints = None;
unsigned int __glutModifierMask = ~0;  /* ~0 implies not in
                                          core input callback. */
int __glutWindowDamaged = 0;

void
glutIdleFunc(GLUTidleCB idleFunc)
{
  __glutIdleFunc = idleFunc;
}

void
glutTimerFunc(unsigned int interval, GLUTtimerCB timerFunc, int value)
{
  GLUTtimer *timer;
  struct timeval now;

  if (!timerFunc)
    return;

  if (freeTimerList) {
    timer = freeTimerList;
    freeTimerList = timer->next;
  } else {
    timer = (GLUTtimer *) malloc(sizeof(GLUTtimer));
    if (!timer)
      __glutFatalError("out of memory.");
  }

#ifdef WIN32
  timer->id = SetTimer( NULL, 0, interval, NULL );
#endif
  timer->func = timerFunc;
#ifdef __vms
  /* VMS time is expressed in units of 100 ns */
  timer->timeout.val = interval * TICKS_PER_MILLISECOND;
#else
  timer->timeout.tv_sec = (int) interval / 1000;
  timer->timeout.tv_usec = (int) (interval % 1000) * 1000;
#endif
  timer->value = value;
  timer->next = NULL;
  GETTIMEOFDAY(&now);
  ADD_TIME(timer->timeout, timer->timeout, now);
#ifdef unix
  prevptr = &__glutTimerList;
  other = *prevptr;
  while (other && IS_AFTER(other->timeout, timer->timeout)) {
    prevptr = &other->next;
    other = *prevptr;
  }
  timer->next = other;
#ifdef SUPPORT_FORTRAN
  __glutNewTimer = timer;  /* for Fortran binding! */
#endif
  *prevptr = timer;
#elif defined( WIN32 )
  timer->next = __glutTimerList;
  __glutTimerList = timer;
#endif
}

#ifdef unix
static void
handleTimeouts(void)
{
  struct timeval now;
  GLUTtimer *timer;

  if (__glutTimerList) {
    GETTIMEOFDAY(&now);
    while (IS_AT_OR_AFTER(__glutTimerList->timeout, now)) {
      timer = __glutTimerList;
      timer->func(timer->value);
      __glutTimerList = timer->next;
      timer->next = freeTimerList;
      freeTimerList = timer;
      if (!__glutTimerList)
        break;
    }
  }
}
#elif defined( WIN32 )
static void
handleTimeouts( unsigned int id )
{
  GLUTtimer *timer, *prev;

	KillTimer( NULL, id );
	prev = NULL;
    timer = __glutTimerList;

  	while ( timer && timer->id != id )
	{
		prev = timer;
		timer = timer->next;
	}

	if ( timer )
	{
		if ( prev )
			prev->next = timer->next;
		else
      		__glutTimerList = timer->next;
      	timer->next = freeTimerList;
      	freeTimerList = timer;
      	(*timer->func) (timer->value);
	}
}
#endif

void
__glutPutOnWorkList(GLUTwindow * window, int workMask)
{
  if (window->workMask) {
    /* Already on list; just OR in new workMask. */
    window->workMask |= workMask;
  } else {
    /* Update work mask and add to window work list. */
    window->workMask = workMask;
    window->prevWorkWin = __glutWindowWorkList;
    __glutWindowWorkList = window;
  }
}

void
__glutPostRedisplay(GLUTwindow * window, int layerMask)
{
  int shown = (layerMask & (GLUT_REDISPLAY_WORK | GLUT_REPAIR_WORK)) ?
    window->shownState : window->overlay->shownState;

  /* Post a redisplay if the window is visible (or the
     visibility of the window is unknown, ie. window->visState
     == -1) _and_ the layer is known to be shown. */
  if (window->visState != GLUT_HIDDEN && window->visState != GLUT_FULLY_COVERED && shown)
    __glutPutOnWorkList(window, layerMask);
}

/* CENTRY */
void
glutPostRedisplay(void)
{
  __glutPostRedisplay(__glutCurrentWindow, GLUT_REDISPLAY_WORK);
}

/* ENDCENTRY */
static GLUTeventParser *eventParserList = NULL;

/* __glutRegisterEventParser allows another module to register
   to intercept X events types not otherwise acted on by the
   GLUT processEventsAndTimeouts routine.  The X Input
   extension support code uses an event parser for handling X
   Input extension events.  */

void
__glutRegisterEventParser(GLUTeventParser * parser)
{
  parser->next = eventParserList;
  eventParserList = parser;
}

static void
markWindowHidden(GLUTwindow * window)
{
  if (GLUT_HIDDEN != window->visState) {
    GLUTwindow *child;

    if (window->windowStatus) {
      window->visState = GLUT_HIDDEN;
      __glutSetWindow(window);
      window->windowStatus(GLUT_HIDDEN);
    }
    /* An unmap is only reported on a single window; its
       descendents need to know they are no longer visible. */
    child = window->children;
    while (child) {
      markWindowHidden(child);
      child = child->siblings;
    }
  }
}

static void
purgeStaleWindow(Window win)
{
  GLUTstale **pEntry = &__glutStaleWindowList;
  GLUTstale *entry = __glutStaleWindowList;

  /* Tranverse singly-linked stale window list look for the
     window ID. */
  while (entry) {
    if (entry->win == win) {
      /* Found it; delete it. */
      *pEntry = entry->next;
      free(entry);
      return;
    } else {
      pEntry = &entry->next;
      entry = *pEntry;
    }
  }
}

#ifdef unix
static void
processEventsAndTimeouts(void)
{
  XEvent event, ahead;
  GLUTwindow *window;
  int width, height;
  GLUTeventParser *parser;

  do {
    XNextEvent(__glutDisplay, &event);
    switch (event.type) {
    case MappingNotify:
      XRefreshKeyboardMapping((XMappingEvent *) & event);
      break;
    case ConfigureNotify:
      window = __glutGetWindow(event.xconfigure.window);
      if (window) {
        if (window->win != event.xconfigure.window) {
          /* Ignore ConfigureNotify sent to the overlay planes.
             GLUT could get here because overlays select for
             StructureNotify events to receive DestroyNotify. */
          break;
        }
        width = event.xconfigure.width;
        height = event.xconfigure.height;
        if (width != window->width || height != window->height) {
          if (window->overlay) {
            XResizeWindow(__glutDisplay, window->overlay->win, width, height);
          }
          window->width = width;
          window->height = height;
          __glutSetWindow(window);
          /* Do not execute OpenGL out of sequence with respect
             to the XResizeWindow request! */
          glXWaitX();
          window->reshape(width, height);
          window->forceReshape = False;
          /* A reshape should be considered like posting a
             redisplay; this is necessary for the "Mesa
             glXSwapBuffers to repair damage" hack to operate
             correctly.  Without it, there's not an initial
             back buffer render from which to blit from when
             damage happens to the window. */
          __glutPostRedisplay(window, GLUT_REDISPLAY_WORK);
        }
      }
      break;
    case Expose:
      /* compress expose events */
      while (XEventsQueued(__glutDisplay, QueuedAfterReading)
        > 0) {
        XPeekEvent(__glutDisplay, &ahead);
        if (ahead.type != Expose ||
          ahead.xexpose.window != event.xexpose.window)
          break;
        XNextEvent(__glutDisplay, &event);
      }
      if (event.xexpose.count == 0) {
        GLUTmenu *menu;

        if (__glutMappedMenu &&
          (menu = __glutGetMenu(event.xexpose.window))) {
          __glutPaintMenu(menu);
        } else {
          window = __glutGetWindow(event.xexpose.window);
          if (window) {
            if (window->win == event.xexpose.window) {
              __glutPostRedisplay(window, GLUT_REPAIR_WORK);
            } else if (window->overlay && window->overlay->win == event.xexpose.window) {
              __glutPostRedisplay(window, GLUT_OVERLAY_REPAIR_WORK);
            }
          }
        }
      } else {
        /* there are more exposes to read; wait to redisplay */
      }
      break;
    case ButtonPress:
    case ButtonRelease:
      if (__glutMappedMenu && event.type == ButtonRelease
        && mappedMenuButton == event.xbutton.button) {
        /* Menu is currently popped up and its button is
           released. */
        __glutFinishMenu(event.xbutton.window, event.xbutton.x, event.xbutton.y);
      } else {
        window = __glutGetWindow(event.xbutton.window);
        if (window) {
          GLUTmenu *menu;

          menu = __glutGetMenuByNum(
            window->menu[event.xbutton.button - 1]);
          if (menu) {
            if (event.type == ButtonPress && !__glutMappedMenu) {
              __glutStartMenu(menu, window,
                event.xbutton.x_root, event.xbutton.y_root,
                event.xbutton.x, event.xbutton.y);
              mappedMenuButton = event.xbutton.button;
            } else {
              /* Ignore a release of a button with a menu
                 attatched to it when no menu is popped up, or
                 ignore a press when another menu is already
                 popped up. */
            }
          } else if (window->mouse) {
            __glutSetWindow(window);
            __glutModifierMask = event.xbutton.state;
            window->mouse(event.xbutton.button - 1,
              event.type == ButtonRelease ?
              GLUT_UP : GLUT_DOWN,
              event.xbutton.x, event.xbutton.y);
            __glutModifierMask = ~0;
          } else {
            /* Stray mouse events.  Ignore. */
          }
        } else {
          /* Window might have been destroyed and all the 
             events for the window may not yet be received. */
        }
      }
      break;
    case MotionNotify:
      if (!__glutMappedMenu) {
        window = __glutGetWindow(event.xmotion.window);
        if (window) {
          /* If motion function registered _and_ buttons held *
             down, call motion function...  */
          if (window->motion && event.xmotion.state &
            (Button1Mask | Button2Mask | Button3Mask)) {
            __glutSetWindow(window);
            window->motion(event.xmotion.x, event.xmotion.y);
          }
          /* If passive motion function registered _and_
             buttons not held down, call passive motion
             function...  */
          else if (window->passive &&
              ((event.xmotion.state &
                  (Button1Mask | Button2Mask | Button3Mask)) ==
              0)) {
            __glutSetWindow(window);
            window->passive(event.xmotion.x,
              event.xmotion.y);
          }
        }
      } else {
        /* Motion events are thrown away when a pop up menu is
           active. */
      }
      break;
    case KeyPress:
      window = __glutGetWindow(event.xkey.window);
      if (!window) {
        break;
      }
      if (window->keyboard) {
        char tmp[1];
        int rc;

        rc = XLookupString(&event.xkey, tmp, sizeof(tmp),
          NULL, NULL);
        if (rc) {
          __glutSetWindow(window);
          __glutModifierMask = event.xkey.state;
          window->keyboard(tmp[0],
            event.xkey.x, event.xkey.y);
          __glutModifierMask = ~0;
          break;
        }
      }
      if (window->special) {
        KeySym ks;
        int key;

        ks = XLookupKeysym((XKeyEvent *) & event, 0);
        /* XXX Verbose, but makes no assumptions about keysym
           layout. */
        switch (ks) {
/* *INDENT-OFF* */
          /* function keys */
          case XK_F1:    key = GLUT_KEY_F1; break;
          case XK_F2:    key = GLUT_KEY_F2; break;
          case XK_F3:    key = GLUT_KEY_F3; break;
          case XK_F4:    key = GLUT_KEY_F4; break;
          case XK_F5:    key = GLUT_KEY_F5; break;
          case XK_F6:    key = GLUT_KEY_F6; break;
          case XK_F7:    key = GLUT_KEY_F7; break;
          case XK_F8:    key = GLUT_KEY_F8; break;
          case XK_F9:    key = GLUT_KEY_F9; break;
          case XK_F10:   key = GLUT_KEY_F10; break;
          case XK_F11:   key = GLUT_KEY_F11; break;
          case XK_F12:   key = GLUT_KEY_F12; break;
          /* directional keys */
          case XK_Left:  key = GLUT_KEY_LEFT; break;
          case XK_Up:    key = GLUT_KEY_UP; break;
          case XK_Right: key = GLUT_KEY_RIGHT; break;
          case XK_Down:  key = GLUT_KEY_DOWN; break;
/* *INDENT-ON* */

        case XK_Prior:
          /* XK_Prior same as X11R6's XK_Page_Up */
          key = GLUT_KEY_PAGE_UP;
          break;
        case XK_Next:
          /* XK_Next same as X11R6's XK_Page_Down */
          key = GLUT_KEY_PAGE_DOWN;
          break;
        case XK_Home:
          key = GLUT_KEY_HOME;
          break;
        case XK_End:
#ifdef __hpux
        case XK_Select:
#endif
          key = GLUT_KEY_END;
          break;
        case XK_Insert:
#ifdef __hpux
        case XK_InsertChar:
#endif
          key = GLUT_KEY_INSERT;
          break;
#ifdef __hpux
        case XK_DeleteChar:
          /* The Delete character is really an ASCII key. */
          __glutSetWindow(window);
          window->keyboard(127,  /* ASCII Delete character. */
            event.xkey.x, event.xkey.y);
          goto skip;
#endif
        default:
          goto skip;
        }
        __glutSetWindow(window);
        __glutModifierMask = event.xkey.state;
        window->special(key, event.xkey.x, event.xkey.y);
        __glutModifierMask = ~0;
      skip:;
      }
      break;
    case EnterNotify:
    case LeaveNotify:
      if (event.xcrossing.mode != NotifyNormal ||
        event.xcrossing.detail == NotifyNonlinearVirtual ||
        event.xcrossing.detail == NotifyVirtual) {

        /* Careful to ignore Enter/LeaveNotify events that come
           from the pop-up menu pointer grab and ungrab.  Also,
           ignore "virtual" Enter/LeaveNotify events since they
           represent the pointer passing through the window
           hierarchy without actually entering or leaving the
           actual real estate of a window.  */

        break;
      }
      if (__glutMappedMenu) {
        GLUTmenuItem *item;
        int num;

        item = __glutGetMenuItem(__glutMappedMenu,
          event.xcrossing.window, &num);
        if (item) {
          __glutMenuItemEnterOrLeave(item, num, event.type);
          break;
        }
      }
      window = __glutGetWindow(event.xcrossing.window);
      if (window) {
        if (window->entry) {
          if (event.type == EnterNotify) {

            /* With overlays established, X can report two
               enter events for both the overlay and normal
               plane window. Do not generate a second enter
               callback if we reported one without an
               intervening leave. */

            if (window->entryState != EnterNotify) {
              int num = window->num;
              Window xid = window->win;

              window->entryState = EnterNotify;
              __glutSetWindow(window);
              window->entry(GLUT_ENTERED);

              if (__glutMappedMenu) {

                /* Do not generate any passive motion events
                   when menus are in use. */

              } else {

                /* An EnterNotify event can result in a
                   "compound" callback if a passive motion
                   callback is also registered. In this case,
                   be a little paranoid about the possibility
                   the window could have been destroyed in the
                   entry callback. */

                window = __glutWindowList[num];
                if (window && window->passive && window->win == xid) {
                  __glutSetWindow(window);
                  window->passive(event.xcrossing.x, event.xcrossing.y);
                }
              }
            }
          } else {
            if (window->entryState != LeaveNotify) {

              /* When an overlay is established for a window
                 already mapped and with the pointer in it, the
                 X server will generate a leave/enter event pair
                 as the pointer leaves (without moving) from the
                 normal plane X window to the newly mapped
                 overlay  X window (or vice versa). This
                 enter/leave pair should not be reported to the
                 GLUT program since the pair is a consequence of
                 creating (or destroying) the overlay, not an
                 actual leave from the GLUT window. */

              if (XEventsQueued(__glutDisplay, QueuedAfterReading)) {
                XPeekEvent(__glutDisplay, &ahead);
                if (ahead.type == EnterNotify &&
                  __glutGetWindow(ahead.xcrossing.window) == window) {
                  XNextEvent(__glutDisplay, &event);
                  break;
                }
              }
              window->entryState = LeaveNotify;
              __glutSetWindow(window);
              window->entry(GLUT_LEFT);
            }
          }
        } else if (window->passive) {
          __glutSetWindow(window);
          window->passive(event.xcrossing.x, event.xcrossing.y);
        }
      }
      break;
    case UnmapNotify:
      /* MapNotify events are not needed to maintain visibility
         state since VisibilityNotify events will be delivered
         when a window becomes visible from mapping.  However,
         VisibilityNotify events are not delivered when a window
         is unmapped (for the window or its children). */
      window = __glutGetWindow(event.xunmap.window);
      if (window) {
        if (window->win != event.xconfigure.window) {
          /* Ignore UnmapNotify sent to the overlay planes.
             GLUT could get here because overlays select for
             StructureNotify events to receive DestroyNotify. */
          break;
        }
        markWindowHidden(window);
      }
      break;
    case VisibilityNotify:
      window = __glutGetWindow(event.xvisibility.window);
      if (window) {
	/* VisibilityUnobscured+1 = GLUT_FULLY_RETAINED,
	   VisibilityPartiallyObscured+1 = GLUT_PARTIALLY_RETAINED,
	   VisibilityFullyObscured+1 = GLUT_FULLY_COVERED. */
        int visState = event.xvisibility.state + 1;

        if (visState != window->visState) {
          if (window->windowStatus) {
            window->visState = visState;
            __glutSetWindow(window);
            window->windowStatus(visState);
          }
        }
      }
      break;
    case ClientMessage:
      if (event.xclient.data.l[0] == __glutWMDeleteWindow)
        exit(0);
      break;
    case DestroyNotify:
      purgeStaleWindow(event.xdestroywindow.window);
      break;
    case CirculateNotify:
    case CreateNotify:
    case GravityNotify:
    case ReparentNotify:
      /* Uninteresting to GLUT (but possible for GLUT to
         receive). */
      break;
    default:
      /* Pass events not directly handled by the GLUT main
         event loop to any event parsers that have been
         registered.  In this way, X Input extension events are
         passed to the correct handler without forcing all GLUT
         programs to support X Input event handling. */
      parser = eventParserList;
      while (parser) {
        if (parser->func(&event))
          break;
        parser = parser->next;
      }
      break;
    }
    if (__glutTimerList)
      handleTimeouts();
  }
  while (XPending(__glutDisplay));
}

static void
waitForSomething(void)
{
#ifdef __vms
  static struct timeval zerotime =
  {0};
  unsigned int timer_efn;
#define timer_id 'glut' /* random :-) number */
  unsigned int wait_mask;
#else
  static struct timeval zerotime =
  {0, 0};
  fd_set fds;
#endif
  struct timeval now, timeout, waittime;
  int rc;

  /* flush X protocol since XPending does not do this
     implicitly */
  XFlush(__glutDisplay);
  if (XPending(__glutDisplay)) {
    /* It is possible (but quite rare) that XFlush may have
       needed to wait for a writable X connection file
       descriptor, and in the process, may have had to read off
       X protocol from the file descriptor. If XPending is true,
       this case occured and we should avoid waiting in select
       since X protocol buffered within Xlib is due to be
       processed and potentially no more X protocol is on the
       file descriptor, so we would risk waiting improperly in
       select. */
    goto immediatelyHandleXinput;
  }
#ifdef __vms
  timeout = __glutTimerList->timeout;
  GETTIMEOFDAY(&now);
  wait_mask = 1 << (__glutConnectionFD & 31);
  if (IS_AFTER(now, timeout)) {
    /* We need an event flag for the timer. */
    /* XXX The `right' way to do this is to use LIB$GET_EF, but
       since it needs to be in the same cluster as the EFN for
       the display, we will have hack it. */
    timer_efn = __glutConnectionFD - 1;
    if ((timer_efn / 32) != (__glutConnectionFD / 32)) {
      timer_efn = __glutConnectionFD + 1;
    }
    rc = SYS$CLREF(timer_efn);
    rc = SYS$SETIMR(timer_efn, &timeout, NULL, timer_id, 0);
    wait_mask |= 1 << (timer_efn & 31);
  } else {
    timer_efn = 0;
  }
  rc = SYS$WFLOR(__glutConnectionFD, wait_mask);
  if (timer_efn != 0 && SYS$CLREF(timer_efn) == SS$_WASCLR) {
    rc = SYS$CANTIM(timer_id, PSL$C_USER);
  }
  /* XXX There does not seem to be checking of "rc" in the code
     above.  Can any of the SYS$ routines above fail? */
#else /* not vms */
  FD_ZERO(&fds);
  FD_SET(__glutConnectionFD, &fds);
  timeout = __glutTimerList->timeout;
  GETTIMEOFDAY(&now);
  if (IS_AFTER(now, timeout)) {
    TIMEDELTA(waittime, timeout, now);
  } else {
    waittime = zerotime;
  }
  rc = select(__glutConnectionFD + 1, &fds,
    NULL, NULL, &waittime);
  if (rc < 0 && errno != EINTR)
    __glutFatalError("select error.");
#endif /* not vms */
  /* Without considering the cause of select unblocking, check
     for pending X events and handle any timeouts (by calling
     processEventsAndTimeouts).  We always look for X events
     even if select returned with 0 (indicating a timeout);
     otherwise we risk starving X event processing by continous 

     timeouts. */
  if (XPending(__glutDisplay)) {
  immediatelyHandleXinput:
    processEventsAndTimeouts();
  } else {
    if (__glutTimerList)
      handleTimeouts();
  }
}

static void
idleWait(void)
{
  if (XPending(__glutDisplay)) {
    processEventsAndTimeouts();
  } else {
    if (__glutTimerList)
      handleTimeouts();
  }
  /* Make sure idle func still exists! */
  if (__glutIdleFunc)
    __glutIdleFunc();
}

#elif defined( WIN32 )

LONG WINAPI
__glutProcessEvents(HWND hWnd, UINT message, DWORD wParam, LONG lParam)
{
    PAINTSTRUCT paint;
    HDC hdc;
	RECT winpos;
	POINT mousepos;
	int x, y, key, i;
	POINT screencoor;
	GLUTwindow *window;
	int button;


	window = __glutCurrentWindow;

	switch (message) {
		case WM_SETCURSOR:
			SetCursor(NULL);
			break;

		case WM_MOVE:
			if(window->ctx && window->renderCtx)
				if(!grSstControl(GR_CONTROL_MOVE))
				{
					PostMessage(hWnd,WM_CLOSE,0,0);
					return(0);
				}
			break;

		case WM_DISPLAYCHANGE:
		case WM_SIZE:
			if(window->ctx && window->renderCtx)
				if(!grSstControl(GR_CONTROL_RESIZE))
				{
					PostMessage(hWnd,WM_CLOSE,0,0);
					return(0);
				}
			break;

		case WM_ACTIVATE:
			if(window->ctx && window->renderCtx)
			{
				WORD	fActive = LOWORD(wParam);
				BOOL	fMinimized = (BOOL) HIWORD(wParam);
				
				if((fActive == WA_INACTIVE) || fMinimized)
					grSstControl(GR_CONTROL_DEACTIVATE);
				else
					grSstControl(GR_CONTROL_ACTIVATE);
			}
			break;

   	  case WM_MOUSEMOVE:
        if (wParam & MK_LBUTTON || wParam & MK_MBUTTON || wParam & MK_RBUTTON) 
		{
			if (window->motion)
			{
				__glutSetWindow( window );
            	(*window->motion)(LOWORD(lParam), HIWORD(lParam));
			}
		}
		else if (window->passive)
		{
			__glutSetWindow( window );
            (*window->passive)(LOWORD(lParam), HIWORD(lParam));
		}
        return (0);

      case WM_LBUTTONDOWN:
      case WM_MBUTTONDOWN:
      case WM_RBUTTONDOWN:
		// get any modifiers
		__glutModifierMask = 0;
		if(GetKeyState(VK_CONTROL) & 0x8000)
			__glutModifierMask |= ControlMask;
		if(GetKeyState(VK_SHIFT) & 0x8000)
			__glutModifierMask |= ShiftMask;
		if(GetKeyState(VK_MENU) & 0x8000)
			__glutModifierMask |= AltMask;

		switch( message )
		{
		  case WM_LBUTTONDOWN: button = GLUT_LEFT_BUTTON; break;
		  case WM_MBUTTONDOWN: button = GLUT_MIDDLE_BUTTON; break;
		  case WM_RBUTTONDOWN: button = GLUT_RIGHT_BUTTON; break;
		}

		if (window->mouse)
        {
			__glutSetWindow( window );
            (*window->mouse)(button, GLUT_DOWN,
            	LOWORD(lParam), HIWORD(lParam));
        }
		// reset the modifier key mask
		__glutModifierMask = (unsigned int)~0;
        return (0);

      case WM_LBUTTONUP:
      case WM_MBUTTONUP:
      case WM_RBUTTONUP:
		// get any modifiers
		__glutModifierMask = 0;
		if(GetKeyState(VK_CONTROL) & 0x8000)
			__glutModifierMask |= ControlMask;
		if(GetKeyState(VK_SHIFT) & 0x8000)
			__glutModifierMask |= ShiftMask;
		if(GetKeyState(VK_MENU) & 0x8000)
			__glutModifierMask |= AltMask;

		switch( message )
		{
		  case WM_LBUTTONUP: button = GLUT_LEFT_BUTTON; break;
		  case WM_MBUTTONUP: button = GLUT_MIDDLE_BUTTON; break;
		  case WM_RBUTTONUP: button = GLUT_RIGHT_BUTTON; break;
		}
		if (window->mouse)
        {
			__glutSetWindow( window );
            (*window->mouse)(button, GLUT_UP,
            	LOWORD(lParam), HIWORD(lParam));
		}
		// reset the modifier key mask
		__glutModifierMask = (unsigned int)~0;
        return (0);

	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
		// get any modifiers
		__glutModifierMask = 0;
		if(GetKeyState(VK_CONTROL) & 0x8000)
			__glutModifierMask |= ControlMask;
		if(GetKeyState(VK_SHIFT) & 0x8000)
			__glutModifierMask |= ShiftMask;
		if(GetKeyState(VK_MENU) & 0x8000)
			__glutModifierMask |= AltMask;

        switch (wParam) 
        {
			case VK_F1: key = GLUT_KEY_F1; break;
			case VK_F2: key = GLUT_KEY_F2; break;
			case VK_F3: key = GLUT_KEY_F3; break;
			case VK_F4: key = GLUT_KEY_F4; break;
			case VK_F5: key = GLUT_KEY_F5; break;
			case VK_F6: key = GLUT_KEY_F6; break;
			case VK_F7: key = GLUT_KEY_F7; break;
			case VK_F8: key = GLUT_KEY_F8; break;
			case VK_F9: key = GLUT_KEY_F9; break;
			case VK_F10: key = GLUT_KEY_F10; break;
			case VK_F11: key = GLUT_KEY_F11; break;
			case VK_F12: key = GLUT_KEY_F12; break;
			case VK_LEFT: key = GLUT_KEY_LEFT; break;
			case VK_UP: key = GLUT_KEY_UP; break;
			case VK_RIGHT: key = GLUT_KEY_RIGHT; break;
			case VK_DOWN: key = GLUT_KEY_DOWN; break;
			case VK_PRIOR: key = GLUT_KEY_PAGE_UP; break;
			case VK_NEXT: key = GLUT_KEY_PAGE_DOWN; break;
			case VK_HOME: key = GLUT_KEY_HOME; break;
			case VK_END: key = GLUT_KEY_END; break;
			case VK_INSERT: key = GLUT_KEY_INSERT; break;

			default:
			  key = FALSE;
			  break;
        }

        if (key && window->special)
        {
			__glutSetWindow( window );
		  GetWindowRect(window->win, &winpos);
		  GetCursorPos(&mousepos);
		  x = mousepos.x - winpos.left;
		  y = mousepos.y - winpos.top;
		  
		  (*window->special)(key, x, y);
        }
		__glutModifierMask = (unsigned int)~0;
        return (0);

      case WM_SYSCHAR:
      case WM_CHAR:
		// get any modifiers
		__glutModifierMask = 0;
		if(GetKeyState(VK_CONTROL) & 0x8000)
			__glutModifierMask |= ControlMask;
		if(GetKeyState(VK_SHIFT) & 0x8000)
			__glutModifierMask |= ShiftMask;
		if(GetKeyState(VK_MENU) & 0x8000)
			__glutModifierMask |= AltMask;
		key = wParam;

        if (key && window->keyboard)
        {
			__glutSetWindow( window );
		  GetWindowRect(window->win, &winpos);
		  GetCursorPos(&mousepos);
		  x = mousepos.x - winpos.left;
		  y = mousepos.y - winpos.top;
		  
		  (*window->keyboard)((unsigned char)key, x, y);
        }
		__glutModifierMask = (unsigned int)~0;
        return (0);

	  case WM_SHOWWINDOW:
		__glutSetWindow ( window );
		window->visState = wParam;

		if (window->visState)
		{
			window->shownState = SW_SHOWNORMAL;
			if (window->visibility)
            	(*window->visibility)(GLUT_VISIBLE);
		}
		else
		{
			window->shownState = SW_HIDE;
			if (window->visibility)
            	(*window->visibility)(GLUT_NOT_VISIBLE);
		}
		glutPostRedisplay();
		return (0);

	  case WM_TIMER:
		  handleTimeouts( wParam );
		  return 0;
    }
    return(DefWindowProc( hWnd, message, wParam, lParam ));
}
#endif

static GLUTwindow **beforeEnd;

static GLUTwindow *
processWindowWorkList(GLUTwindow * window)
{
  int workMask;

  if (window->prevWorkWin)
    window->prevWorkWin = processWindowWorkList(window->prevWorkWin);
  else
    beforeEnd = &window->prevWorkWin;

  /* Capture work mask for work that needs to be done to this
     window, then clear the window's work mask (excepting the
     dummy work bit, see below).  Then, process the captured
     work mask.  This allows callbacks in the processing the
     captured work mask to set the window's work mask for
     subsequent processing. */

  workMask = window->workMask;
  assert((workMask & GLUT_DUMMY_WORK) == 0);

  /* Set the dummy work bit, clearing all other bits, to
     indicate that the window is currently on the window work
     list _and_ that the window's work mask is currently being
     processed.  This convinces __glutPutOnWorkList that this
     window is on the work list still. */
  window->workMask = GLUT_DUMMY_WORK;

  /* Optimization: most of the time, the work to do is a
     redisplay and not these other types of work.  Check for
     the following cases as a group to before checking each one
     individually one by one. This saves about 25 MIPS
     instructions in the common redisplay only case. */
  if (workMask & (GLUT_EVENT_MASK_WORK | GLUT_DEVICE_MASK_WORK |
      GLUT_CONFIGURE_WORK | GLUT_COLORMAP_WORK | GLUT_MAP_WORK)) {
    /* Be sure to set event mask *BEFORE* map window is done. */
    if (workMask & GLUT_EVENT_MASK_WORK) {
    }

    /* Be sure to set device mask *BEFORE* map window is done. */
    if (workMask & GLUT_DEVICE_MASK_WORK) {
      __glutUpdateInputDeviceMaskFunc(window);
    }
    /* Be sure to configure window *BEFORE* map window is done. 
     */
    if (workMask & GLUT_CONFIGURE_WORK) {
      window->desiredConfMask = 0;
    }
    /* Be sure to establish the colormaps *BEFORE* map window
       is done. */
    if (workMask & GLUT_COLORMAP_WORK) {
//      __glutEstablishColormapsProperty(window);
    }
    if (workMask & GLUT_MAP_WORK) {
    }
  }
  if (workMask & (GLUT_REDISPLAY_WORK | GLUT_OVERLAY_REDISPLAY_WORK | GLUT_REPAIR_WORK | GLUT_OVERLAY_REPAIR_WORK)) {
    if (window->forceReshape) {
      /* Guarantee that before a display callback is generated
         for a window, a reshape callback must be generated. */
      __glutSetWindow(window);
      window->reshape(window->width, window->height);
      window->forceReshape = False;

      /* Setting the redisplay bit on the first reshape is
         necessary to make the "Mesa glXSwapBuffers to repair
         damage" hack operate correctly.  Without indicating a
         redisplay is necessary, there's not an initial back
         buffer render from which to blit from when damage
         happens to the window. */
      workMask |= GLUT_REDISPLAY_WORK;
    }
    /* The code below is more involved than otherwise necessary
       because it is paranoid about the overlay or entire window
       being removed or destroyed in the course of the callbacks.
       Notice how the global __glutWindowDamaged is used to record
       the layers' damage status.  See the code in glutLayerGet for
       how __glutWindowDamaged is used. The  point is to not have to
       update the "damaged" field after  the callback since the
       window (or overlay) may be destroyed (or removed) when the
       callback returns. */

    if (window->overlay && window->overlay->display) {
      int num = window->num;
      Window xid = window->overlay ? window->overlay->win : None;

      /* If an overlay display callback is registered, we
         differentiate between a redisplay needed for the
         overlay and/or normal plane.  If there is no overlay
         display callback registered, we simply use the
         standard display callback. */

      if (workMask & (GLUT_REDISPLAY_WORK | GLUT_REPAIR_WORK)) {
        if (__glutMesaSwapHackSupport) {
          if (window->usedSwapBuffers) {
            if ((workMask & (GLUT_REPAIR_WORK | GLUT_REDISPLAY_WORK)) == GLUT_REPAIR_WORK) {
              fxMesaSwapBuffers();
              goto skippedDisplayCallback1;
            }
          }
        }
        /* Render to normal plane. */
        window->renderWin = window->win;
        window->renderCtx = window->ctx;
        __glutWindowDamaged = (workMask & GLUT_REPAIR_WORK);
        __glutSetWindow(window);
        window->usedSwapBuffers = 0;
        window->display();
        __glutWindowDamaged = 0;

      skippedDisplayCallback1:;
      }
      if (workMask & (GLUT_OVERLAY_REDISPLAY_WORK | GLUT_OVERLAY_REPAIR_WORK)) {
        window = __glutWindowList[num];
        if (window && window->overlay &&
          window->overlay->win == xid && window->overlay->display) {

          /* Render to overlay. */
          window->renderWin = window->overlay->win;
          window->renderCtx = window->overlay->ctx;
          __glutWindowDamaged = (workMask & GLUT_OVERLAY_REPAIR_WORK);
          __glutSetWindow(window);
          window->overlay->display();
          __glutWindowDamaged = 0;
        } else {
          /* Overlay may have since been destroyed or the
             overlay callback may have been disabled during
             normal display callback. */
        }
      }
    } else {
      if (__glutMesaSwapHackSupport) {
        if (!window->overlay && window->usedSwapBuffers) {
          if ((workMask & (GLUT_REPAIR_WORK | GLUT_REDISPLAY_WORK)) == GLUT_REPAIR_WORK) {
              fxMesaSwapBuffers();
            goto skippedDisplayCallback2;
          }
        }
      }
      /* Render to normal plane (and possibly overlay). */
      __glutWindowDamaged = (workMask & (GLUT_OVERLAY_REPAIR_WORK | GLUT_REPAIR_WORK));
      __glutSetWindow(window);
      window->usedSwapBuffers = 0;
      window->display();
      __glutWindowDamaged = 0;

    skippedDisplayCallback2:;
    }
  }
  /* Combine workMask with window->workMask to determine what
     finish and debug work there is. */
  workMask |= window->workMask;

  if (workMask & GLUT_FINISH_WORK) {
    __glutSetWindow(window);
    glFinish();
  }
  if (workMask & GLUT_DEBUG_WORK) {
    GLenum error;

    __glutSetWindow(window);
    while ((error = glGetError()) != GL_NO_ERROR)
      __glutWarning("GL error: %s", gluErrorString(error));
  }
  /* Strip out dummy, finish, and debug work bits. */
  window->workMask &= ~(GLUT_DUMMY_WORK | GLUT_FINISH_WORK | GLUT_DEBUG_WORK);
  if (window->workMask) {
    /* Leave on work list. */
    return window;
  } else {
    /* Remove current window from work list. */
    return window->prevWorkWin;
  }
}
/* CENTRY */
void
glutMainLoop(void)
{
  if (!__glutDisplay)
    __glutFatalUsage("main loop entered with out X connection.");
  if (!__glutWindowListSize)
    __glutFatalUsage(
      "main loop entered with no windows created.");
#ifdef unix
  for (;;) {
    if (__glutWindowWorkList) {
      GLUTwindow *remainder, *work;

      work = __glutWindowWorkList;
      __glutWindowWorkList = NULL;
      if (work) {
        remainder = processWindowWorkList(work);
        if (remainder) {
          *beforeEnd = __glutWindowWorkList;
          __glutWindowWorkList = remainder;
        }
      }
    }
    if (__glutIdleFunc || __glutWindowWorkList) {
      idleWait();
    } else {
      if (__glutTimerList) {
        waitForSomething();
      } else {
        processEventsAndTimeouts();
      }
    }
  }
#elif defined( WIN32 )
	for (;;)
	{
		if (__glutWindowWorkList) 
		{
			GLUTwindow *remainder, *work;

			work = __glutWindowWorkList;
			__glutWindowWorkList = NULL;
			if (work) 
			{
				remainder = processWindowWorkList(work);
				if (remainder) 
				{
					*beforeEnd = __glutWindowWorkList;
					__glutWindowWorkList = remainder;
				}
			}
		}
		if (__glutIdleFunc || __glutWindowWorkList) 
		{
			MSG Message;

			/*
			 *  Process all pending messages
			 */
		  
			while (PeekMessage(&Message, NULL, 0, 0, PM_NOREMOVE) == TRUE)
			{
				if (GetMessage(&Message, NULL, 0, 0) )
				{
					if ( Message.message == WM_TIMER )
						handleTimeouts( Message.wParam );
					else
					{
						TranslateMessage(&Message);
						DispatchMessage(&Message);
					}
				}
				else
				{
					/*
					 *  Nothing else to do here, just return
					 */
					return;
				}
			}
	/* 
	 * Make sure idle func still exists!
	 */
			if (__glutIdleFunc)
				__glutIdleFunc();
				
		} 
		else 
		{
			MSG Message;

			/*
			 *  Process all messages in a blocking manner
			 */

			while ( !(__glutIdleFunc || __glutWindowWorkList) )
			{
				if (GetMessage(&Message, NULL, 0, 0) )
				{
					if ( Message.message == WM_TIMER )
						handleTimeouts( Message.wParam );
					else
					{
						TranslateMessage(&Message);
						DispatchMessage(&Message);
					}
				}
				else
				{
					/*
					 *  Nothing else to do here, just return
					 */
					return;
				}
			}
		}
	}
#endif
}
/* ENDCENTRY */
